self.__precacheManifest = (self.__precacheManifest || []).concat([
  {
    "revision": "5cd2ac0d70b427b81981",
    "url": "/css/app.a33d2315.css"
  },
  {
    "revision": "c6f530eafaea3e9bbda2",
    "url": "/css/chunk-vendors.fa0e2d5e.css"
  },
  {
    "revision": "82cf43358c1d5de3b328f77464ef3d42",
    "url": "/d3.min.js"
  },
  {
    "revision": "143146fa24554ae2c5ac0a3982abb952",
    "url": "/fonts/ionicons.143146fa.woff2"
  },
  {
    "revision": "99ac3308dd8ee14f749f51538d0d5b9e",
    "url": "/fonts/ionicons.99ac3308.woff"
  },
  {
    "revision": "d535a25a79fb1365ae814b61e88fae71",
    "url": "/fonts/ionicons.d535a25a.ttf"
  },
  {
    "revision": "a2c4a261a239aa84463dc70e4bac9b9a",
    "url": "/img/ionicons.a2c4a261.svg"
  },
  {
    "revision": "165e0f6443a0fdd989d469d4c4c3842e",
    "url": "/img/logo.165e0f64.png"
  },
  {
    "revision": "eb409c5f0304d7866fb2d1ddb35af11a",
    "url": "/index.html"
  },
  {
    "revision": "5cd2ac0d70b427b81981",
    "url": "/js/app.807afc8a.js"
  },
  {
    "revision": "c6f530eafaea3e9bbda2",
    "url": "/js/chunk-vendors.4d8f257e.js"
  },
  {
    "revision": "b35689f0b52b63e33caaf8a8fcd00a64",
    "url": "/manifest.json"
  },
  {
    "revision": "b6216d61c03e6ce0c9aea6ca7808f7ca",
    "url": "/robots.txt"
  }
]);